/*
				P4-1.c - Data Structure Library					
<Kenneth Marino>   							<4/18/12>
*/

#include <stdio.h>     
#include <stdlib.h>
#include "P4-1.h"

typedef struct NodeHeader {
	struct NodeHeader *next;
	struct NodeHeader *prev;
} NodeHeader;

typedef struct {
	NodeHeader *head;
	NodeHeader *tail;
	int (*Match)(void*, void*);
	int (*LE)(void*, void*);
	int ElementSize;
} ListHead;

// Prototype for new helper function
void RemoveElement(void* MyList, NodeHeader* header);

//Create and return a pointer to a doubly linked list with elements of 
//size ElementSize in bytes. 
void *CreateLL(int ElementSize,int (*Match)(void *key, void *MyElement), int (*LE)(void *MyElement1, void *MyElement2))
{
	// Allocate space for the list head
	ListHead* list = malloc(sizeof(ListHead));

	// Set the fields of the list head
	list->head = NULL;
	list->tail = NULL;
	list->Match = Match;
	list->LE = LE;
	list->ElementSize = ElementSize;

	// Cast it to a void pointer and return it
	return (void*)list;
}

// Delete a linked list given a pointer to that list
void DeleteLL(void* MyList)
{
	// Free the memory
	free(MyList);
}

//Return a pointer to an element for LL *MyList.
void *CreateElement(void *MyList)
{
	// Cast to the list head
	ListHead* list = (ListHead*) MyList;
	
	// Allocate space for the element plus the NodeHeader
	void* element = malloc(sizeof(NodeHeader)+list->ElementSize);
	
	// Cast to NodeHeader and set the prev and next pointers to null
	NodeHeader* header = (NodeHeader*) element;
	header->next = NULL;
	header->prev = NULL;

	// Go to start of data and return the element
	element = (void*)(header+1);
	return element;
}

//Delete and recover memory from element *MyElement.  DeleteElement 
//does not check that *MyElement is not a part of any linked list.
void DeleteElement(void *MyElement)
{
	// Cast to NodeHeader and decrement back to the NodeHeader
	NodeHeader* element = (NodeHeader*)MyElement;
	element = element - 1;
	
	// Free the pointer
	free(element);
}

//Insert *MyElement at the head of linked list *MyList.
void InsertElement(void *MyList, void *MyElement)
{
	// Cast the list to make it useable
	ListHead* list = (ListHead*)MyList;

	// Cast element to a NodeHead, and traverse to the node header
	NodeHeader* header = (NodeHeader*)(MyElement);
	header = header - 1;

	// Set prev as NULL and next as the current head
	header->prev = NULL;
	header->next = list->head;

	// If the list is currently empty, make the element the tail as well
	if (list->head == NULL)	
		list->tail = header;

	// Otherwise set the current head's prev as the element
	else
	{
		list->head->prev = header;
	}
	// Set the new element as the head
	list->head = header;
}

// Insert *MyElement in sorted order in linked list *MyList. 
//MyElement->key should be less than or equal to the key of the next 
//element in the list.
void InsertElementSorted(void *MyList, void *MyElement)
{
	// Cast the list to make it useable and get the LE function
	ListHead* list = (ListHead*)MyList;
	int (*LE)(void*, void*) = list->LE;

	// Cast element to a NodeHead, and traverse to the node header
	NodeHeader* MyElHeader = (NodeHeader*)(MyElement);
	MyElHeader = MyElHeader - 1;

	// Get the head of the list
	NodeHeader *header = list->head;

	// If the list is empty, add the element at the head with s the head and the tail
	if (header == NULL)
	{
		InsertElement(MyList, MyElement);
		return;
	}

	// Get the data inside the header
	void* element = (void*)(header+1);

	// If the element is less than the head, then insert it at the head
	if ((*LE)(MyElement, element) == 1)
	{
		InsertElement(MyList, MyElement);
		return;
	}

	// Go to the next element
	header = header->next;

	// Otherwise go through the list until you get to an an element that is greater than MyElement
	while (header != NULL)
	{
		element = (void*)(header+1);
		if ((*LE)(MyElement, element))
		{
			// Set next and prev of the new node
			MyElHeader->prev = header->prev;
			MyElHeader->next = header;
			
			// Set the next of the last and the previous to the new node
			header->prev->next = MyElHeader;
			header->prev = MyElHeader;			
			return;
		}
		header = header->next;
	}
	
	// If it is greater than all of the elements
	// Set the next and prev of the new node
	MyElHeader->prev = list->tail;
	MyElHeader->next = NULL;
	
	// Set new node as next of old tail, and make it the new tail	
	list->tail->next = MyElHeader;
	list->tail = MyElHeader;
}

//Return a pointer to the first element in list *Mylist whose key 
//matches key.  If (Remove == True) (1), the returned element should 
//be removed from the list.
void *GetElement(void *MyList, void *key, int Remove)
{
	// Cast the list to make it useable and get the Match function
	ListHead* list = (ListHead*)MyList;
	int (*Match)(void*, void*) = list->Match;
	NodeHeader* header;
	void* returnElement = NULL;

	// Traverse the list until we get to the end, or we find the key
	header = list->head;
	while (header != NULL)
	{
		// Get the actual element
		void* element = (void*)(header+1);
		int isMatch = (*Match)(key, element);
		if (isMatch == 1)
		{
			returnElement = element; 		
			break;
		}

		// Go to next header
		header = header->next;
	}

	// If the user asked for it to be removed, removes it
	if (Remove == 1)
	{
		// If the element wasn't found, returns
		if (returnElement == NULL)
			return returnElement;
	
		// Otherwise, removes it from the list
		header = (NodeHeader*)returnElement;
		RemoveElement(MyList, header-1);
	}	

	// Returns the element
	return returnElement;
}

//Return a pointer to the first or last element in the list *MyList.
//If (FL == 0) return a pointer to the first element, else return
//the last element. If (Remove == True) (1), the returned element
//should be removed from the list.
void *GetFLElement(void *MyList, int FL, int Remove)
{
	// Cast the list to make it useable
	ListHead* list = (ListHead*)MyList;
	NodeHeader* header;

	void* element;
	// If FL == 0, set element to the head
	if (FL == 0)
	{
		header = list->head;
		// If the head is NULL, return NULL
		if (header == NULL)
			return NULL;
		element = (void*)(header+1);
	}
	// Otherwise, set element to the tail
	else
	{
		header = list->tail;
		// If the tail is NULL, return NULL
		if (header == NULL)
			return NULL;
		element = (void*)(header+1);
	}

	// If it's a remove, delete the element and remove the references in the list
	if (Remove == 1)
		RemoveElement(MyList, header);

	return element;
}

// Return a pointer to the next(previous) object in the list
// direction == 0 is next, direction == 1 is previous
// return NULL if *MyElement is head(tail) and direction is 
// previous(next)
void *Traverse(void *MyElement, int direction)
{
	// Cast element to a NodeHead, and traverse to the node header
	NodeHeader* header = (NodeHeader*)(MyElement);
	header = header - 1;
	
	// If the direction is next
	if (direction == 0)
	{
		// Return NULL if it's the tail
		if (header->next == NULL)
			return NULL;
		// Else get the next element
		else
			return (void*)((header->next)+1);	
	}
	// If the direction is previous
	else
	{
		// Return NULL if it's the head
		if (header->prev == NULL)
			return NULL;
		// Else get the previous element
		else
			return (void*)((header->prev)+1);
	}
}

// Removes the given element from the linked list
void RemoveElement(void *MyList, NodeHeader* header)
{
	// Cast the list
	ListHead* list = (ListHead*)MyList;

	// If it's the head node, change the head to the next node
	if (header->prev == NULL)
	{
		list->head = header->next;
	}
	// Otherwise set the previous node's next to the node's next
	else
	{
		header->prev->next = header->next;
	}

	// If it's the tail node, change the tail to the previous node
	if (header->next == NULL)
	{
		list->tail = header->prev;
	}		
	// Otherwise set the next node's prev to the node's prev
	else
	{
		header->next->prev = header->prev;
	}
}
